
"""
VulnScan - Advanced Vulnerability Scanner

A comprehensive web application vulnerability scanner with AI-powered detection.
"""

__version__ = "1.0.3"
__author__ = "Gokul Kannan Ganesamoorthy"
__email__ = "gokulkannan.dev@gmail.com"

# Import main components for easier access
from .main import *
from .cli import main as cli_main

__all__ = ["main", "cli_main", "__version__"]



import re
from vulnscan.base.testing import tokenPattern


def longestCommonSubstring(s1, s2):
    m = [[0] * (1 + len(s2)) for i in range(1 + len(s1))]
    longest, x_longest = 0, 0
    for x in range(1, 1 + len(s1)):
        for y in range(1, 1 + len(s2)):
            if s1[x - 1] == s2[y - 1]:
                m[x][y] = m[x - 1][y - 1] + 1
                if m[x][y] > longest:
                    longest = m[x][y]
                    x_longest = x
            else:
                m[x][y] = 0
    return s1[x_longest - longest: x_longest]


def stringToBinary(string):
    return ''.join(format(ord(x), 'b') for x in string)


def strength(string):
    digits = re.findall(r'\d', string)
    lowerAlphas = re.findall(r'[a-z]', string)
    upperAlphas = re.findall(r'[A-Z]', string)
    entropy = len(set(digits + lowerAlphas + upperAlphas))
    if not digits:
        entropy = entropy/2
    return entropy


def isProtected(parsed):
    protected = False
    parsedForms = list(parsed.values())
    for oneForm in parsedForms:
        inputs = oneForm['inputs']
        for inp in inputs:
            name = inp['name']
            kind = inp['type']
            value = inp['value']
            if re.match(tokenPattern, value):
                protected = True
    return protected


def extractHeaders(headers):
    headers = headers.replace('\\n', '\n')
    sorted_headers = {}
    matches = re.findall(r'(.*):\s(.*)', headers)
    for match in matches:
        header = match[0]
        value = match[1]
        try:
            if value[-1] == ',':
                value = value[:-1]
            sorted_headers[header] = value
        except IndexError:
            pass
    return sorted_headers


def getUrl(url, data, GET):
    if GET:
        return url.split('?')[0]
    else:
        return url


def getParams(url, data, GET):
    params = {}
    if GET:
        if '=' in url:
            data = url.split('?')[1]
            if data[:1] == '?':
                data = data[1:]
        else:
            data = ''
    parts = data.split('&')
    for part in parts:
        each = part.split('=')
        try:
            params[each[0]] = each[1]
        except IndexError:
            params = None
    return params


def remove_file(url):
    if url.count('/') > 2:
        replacable = re.search(r'/[^/]*?$', url).group()
        if replacable != '/':
            return url.replace(replacable, '')
        else:
            return url
    else:
        return url



#!/usr/bin/env python3
from vulnscan.main import main
import sys
import os

# Add the current directory to the path so we can import vulnscan
sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))


def main():
    from vulnscan.main import main as run_scan
    run_scan()


if __name__ == "__main__":
    main()



def rangeFinder(tokens):
    digits = set()
    alphabets = set()
    for token in tokens:
        for char in token:
            if char in '0123456789':
                digits.add(char)
            elif char in 'abcdefghijklmnopqrstuvwxyz':
                alphabets.add(char)
    return [list(digits), list(alphabets)]



import time
import random
import warnings
import requests

warnings.filterwarnings('ignore')  # Disable SSL related warnings


def sendRequest(url, data, headers, GET, delay):
    time.sleep(delay)
    user_agents = ['Mozilla/5.0 (X11; Linux i686; rv:60.0) Gecko/20100101 Firefox/60.0',
                   'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.113 Safari/537.36'
                   'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/56.0.2924.87 Safari/537.36 OPR/43.0.2442.991']
    if headers:
        if 'User-Agent' not in headers:
            headers['User-Agent'] = random.choice(user_agents)
    if GET:
        response = requests.get(
            url, params=data, headers=headers, verify=False)
    else:
        response = requests.post(url, data=data, headers=headers, verify=False)
    return response



import numpy as np
from math import floor, log
import scipy.special as spc
import scipy.fftpack as sff
import scipy.stats as sst
from functools import reduce


def sumi(x): return 2 * x - 1


def su(x, y): return x + y


def sus(x): return (x - 0.5) ** 2


def sq(x): return int(x) ** 2


def logo(x): return x * np.log(x)


def pr(u, x):
    if u == 0:
        out = 1.0 * np.exp(-x)
    else:
        out = 1.0 * x * np.exp(2*-x) * (2**-u) * spc.hyp1f1(u + 1, 2, x)
    return out


def stringpart(binin, num):
    blocks = [binin[xs * num:num + xs * num:]
              for xs in range(floor(len(binin) / num))]
    return blocks


def randgen(num):
    '''Spits out a stream of random numbers like '1001001' with the length num'''

    rn = open('/dev/urandom', 'r')
    random_chars = rn.read(num / 2)
    stream = ''
    for char in random_chars:
        c = ord(char)
        for i in range(0, 2):
            stream += str(c >> i & 1)
    return stream


def monobitfrequencytest(binin):
    ''' The focus of the test is the proportion of zeroes and ones for the entire sequence. The purpose of this test is to determine whether that number of ones and zeros in a sequence are approximately the same as would be expected for a truly random sequence. The test assesses the closeness of the fraction of ones to 1/2, that is, the number of ones and zeroes in a sequence should be about the same.'''

    ss = [int(el) for el in binin]
    sc = list(map(sumi, ss))
    sn = reduce(su, sc)
    sobs = np.abs(sn) / np.sqrt(len(binin))
    pval = spc.erfc(sobs / np.sqrt(2))
    return pval


def blockfrequencytest(binin, nu=20):
    ''' The focus of the test is the proportion of zeroes and ones within M-bit blocks. The purpose of this test is to determine whether the frequency of ones is an M-bit block is approximately M/2.'''
    ss = [int(el) for el in binin]
    tt = [1.0 * sum(ss[xs * nu:nu + xs * nu:]) /
          nu for xs in range(floor(len(ss) / nu))]
    uu = list(map(sus, tt))
    chisqr = 4 * nu * reduce(su, uu, 0)
    pval = spc.gammaincc(len(tt) / 2.0, chisqr / 2.0)
    return pval


def runstest(binin):
    ''' The focus of this test is the total number of zero and one runs in the entire sequence, where a run is an uninterrupted sequence of identical bits. A run of length k means that a run consists of exactly k identical bits and is bounded before and after with a bit of the opposite value. The purpose of the runs test is to determine whether the number of runs of ones and zeros of various lengths is as expected for a random sequence. In particular, this test determines whether the oscillation between such substrings is too fast or too slow.'''
    ss = [int(el) for el in binin]
    n = len(binin)
    pi = 1.0 * reduce(su, ss) / n
    vobs = len(binin.replace('0', ' ').split()) + \
        len(binin.replace('1', ' ').split())
    pval = spc.erfc(abs(vobs-2*n*pi*(1-pi)) /
                    (2 * pi * (1 - pi) * np.sqrt(2*n)))
    return pval


def longestrunones8(binin):
    ''' The focus of the test is the longest run of ones within M-bit blocks. The purpose of this test is to determine whether the length of the longest run of ones within the tested sequence is consistent with the length of the longest run of ones that would be expected in a random sequence. Note that an irregularity in the expected length of the longest run of ones implies that there is also an irregularity in the expected length of the longest run of zeroes. Long runs of zeroes were not evaluated separately due to a concern about statistical independence among the tests.'''
    m = 8
    k = 3
    pik = [0.2148, 0.3672, 0.2305, 0.1875]
    blocks = [binin[xs*m:m+xs*m:] for xs in range(len(binin) / m)]
    n = len(blocks)
    # append the string 01 to guarantee the length of 1
    counts1 = [xs+'01' for xs in blocks]
    counts = [xs.replace('0', ' ').split()
              for xs in counts1]  # split into all parts
    counts2 = [list(map(len, xx)) for xx in counts]
    counts4 = [(4 if xx > 4 else xx) for xx in map(max, counts2)]
    freqs = [counts4.count(spi) for spi in [1, 2, 3, 4]]
    chisqr1 = [(freqs[xx]-n*pik[xx])**2/(n*pik[xx]) for xx in range(4)]
    chisqr = reduce(su, chisqr1)
    pval = spc.gammaincc(k / 2.0, chisqr / 2.0)
    return pval


def longestrunones128(binin):  # not well tested yet
    if len(binin) > 128:
        m = 128
        k = 5
        n = len(binin)
        pik = [0.1174, 0.2430, 0.2493, 0.1752, 0.1027, 0.1124]
        blocks = [binin[xs * m:m + xs * m:] for xs in range(len(binin) / m)]
        n = len(blocks)
        counts = [xs.replace('0', ' ').split() for xs in blocks]
        counts2 = [list(map(len, xx)) for xx in counts]
        counts3 = [(1 if xx < 1 else xx) for xx in map(max, counts2)]
        counts4 = [(4 if xx > 4 else xx) for xx in counts3]
        chisqr1 = [(counts4[xx] - n * pik[xx]) ** 2 / (n * pik[xx])
                   for xx in range(len(counts4))]
        chisqr = reduce(su, chisqr1)
        pval = spc.gammaincc(k / 2.0, chisqr / 2.0)
    else:
        print('longestrunones128 failed, too few bits:', len(binin))
        pval = 0
    return pval


def longestrunones10000(binin):  # not well tested yet
    ''' The focus of the test is the longest run of ones within M-bit blocks. The purpose of this test is to determine whether the length of the longest run of ones within the tested sequence is consistent with the length of the longest run of ones that would be expected in a random sequence. Note that an irregularity in the expected length of the longest run of ones implies that there is also an irregularity in the expected length of the longest run of zeroes. Long runs of zeroes were not evaluated separately due to a concern about statistical independence among the tests.'''
    if len(binin) > 128:
        m = 10000
        k = 6
        pik = [0.0882, 0.2092, 0.2483, 0.1933, 0.1208, 0.0675, 0.0727]
        blocks = [binin[xs * m:m + xs * m:]
                  for xs in range(floor(len(binin) / m))]
        n = len(blocks)
        counts = [xs.replace('0', ' ').split() for xs in blocks]
        counts2 = [list(map(len, xx)) for xx in counts]
        counts3 = [(10 if xx < 10 else xx) for xx in map(max, counts2)]
        counts4 = [(16 if xx > 16 else xx) for xx in counts3]
        freqs = [counts4.count(spi) for spi in [10, 11, 12, 13, 14, 15, 16]]
        chisqr1 = [(freqs[xx] - n * pik[xx]) ** 2 / (n * pik[xx])
                   for xx in range(len(freqs))]
        chisqr = reduce(su, chisqr1)
        pval = spc.gammaincc(k / 2.0, chisqr / 2.0)
    else:
        print('longestrunones10000 failed, too few bits:', len(binin))
        pval = 0
    return pval

# test 2.06


def spectraltest(binin):
    '''The focus of this test is the peak heights in the discrete Fast Fourier Transform. The purpose of this test is to detect periodic features (i.e., repetitive patterns that are near each other) in the tested sequence that would indicate a deviation from the assumption of randomness. '''

    n = len(binin)
    ss = [int(el) for el in binin]
    sc = list(map(sumi, ss))
    ft = sff.fft(sc)
    af = abs(ft)[1:floor(n/2)+1:]
    t = np.sqrt(np.log(1/0.05)*n)
    n0 = 0.95*n/2
    n1 = len(np.where(af < t)[0])
    d = (n1 - n0)/np.sqrt(n*0.95*0.05/4)
    pval = spc.erfc(abs(d)/np.sqrt(2))
    return pval


def nonoverlappingtemplatematchingtest(binin, mat="000000001", num=9):
    ''' The focus of this test is the number of occurrences of pre-defined target substrings. The purpose of this test is to reject sequences that exhibit too many occurrences of a given non-periodic (aperiodic) pattern. For this test and for the Overlapping Template Matching test, an m-bit window is used to search for a specific m-bit pattern. If the pattern is not found, the window slides one bit position. For this test, when the pattern is found, the window is reset to the bit after the found pattern, and the search resumes.'''
    n = len(binin)
    m = len(mat)
    M = floor(n/num)
    blocks = [binin[xs*M:M+xs*M:] for xs in range(floor(n/M))]
    counts = [xx.count(mat) for xx in blocks]
    avg = 1.0 * (M-m+1)/2 ** m
    var = M*(2**-m - (2*m-1)*2**(-2*m))
    chisqr = reduce(su, [(xs - avg) ** 2 for xs in counts]) / var
    pval = spc.gammaincc(1.0 * len(blocks) / 2, chisqr / 2)
    return pval


def occurances(string, sub):
    count = start = 0
    while True:
        start = string.find(sub, start)+1
        if start > 0:
            count += 1
        else:
            return count


def overlappingtemplatematchingtest(binin, mat="111111111", num=1032, numi=9):
    ''' The focus of this test is the number of pre-defined target substrings. The purpose of this test is to reject sequences that show deviations from the expected number of runs of ones of a given length. Note that when there is a deviation from the expected number of ones of a given length, there is also a deviation in the runs of zeroes. Runs of zeroes were not evaluated separately due to a concern about statistical independence among the tests. For this test and for the Non-overlapping Template Matching test, an m-bit window is used to search for a specific m-bit pattern. If the pattern is not found, the window slides one bit position. For this test, when the pattern is found, the window again slides one bit, and the search is resumed.'''
    n = len(binin)
    bign = int(n / num)
    m = len(mat)
    lamda = 1.0 * (num - m + 1) / 2 ** m
    eta = 0.5 * lamda
    pi = [pr(i, eta) for i in range(numi)]
    pi.append(1 - reduce(su, pi))
    v = [0 for x in range(numi + 1)]
    blocks = stringpart(binin, num)
    blocklen = len(blocks[0])
    counts = [occurances(i, mat) for i in blocks]
    counts2 = [(numi if xx > numi else xx) for xx in counts]
    for i in counts2:
        v[i] = v[i] + 1
    chisqr = reduce(su, [(v[i]-bign*pi[i]) ** 2 / (bign*pi[i])
                         for i in range(numi + 1)])
    pval = spc.gammaincc(0.5*numi, 0.5*chisqr)
    return pval


def maurersuniversalstatistictest(binin, l=6, q=640):
    ''' The focus of this test is the number of bits between matching patterns. The purpose of the test is to detect whether or not the sequence can be significantly compressed without loss of information. An overly compressible sequence is considered to be non-random.'''
    ru = [
        [0.7326495, 0.690],
        [1.5374383, 1.338],
        [2.4016068, 1.901],
        [3.3112247, 2.358],
        [4.2534266, 2.705],
        [5.2177052, 2.954],
        [6.1962507, 3.125],
        [7.1836656, 3.238],
        [8.1764248, 3.311],
        [9.1723243, 3.356],
        [10.170032, 3.384],
        [11.168765, 3.401],
        [12.168070, 3.410],
        [13.167693, 3.416],
        [14.167488, 3.419],
        [15.167379, 3.421],
    ]
    blocks = [int(li, 2) + 1 for li in stringpart(binin, l)]
    k = len(blocks) - q
    states = [0 for x in range(2**l)]
    for x in range(q):
        states[blocks[x]-1] = x+1
    sumi = 0.0
    for x in range(q, len(blocks)):
        sumi += np.log2((x+1)-states[blocks[x]-1])
        states[blocks[x]-1] = x+1
    fn = sumi / k
    c = 0.7-(0.8/l)+(4+(32.0/l))*((k**(-3.0/l))/15)
    sigma = c*np.sqrt((ru[l-1][1])/k)
    pval = spc.erfc(abs(fn-ru[l-1][0]) / (np.sqrt(2)*sigma))
    return pval


def lempelzivcompressiontest1(binin):
    ''' The focus of this test is the number of cumulatively distinct patterns (words) in the sequence. The purpose of the test is to determine how far the tested sequence can be compressed. The sequence is considered to be non-random if it can be significantly compressed. A random sequence will have a characteristic number of distinct patterns.'''
    i = 1
    j = 0
    n = len(binin)
    mu = 69586.25
    sigma = 70.448718
    words = []
    while (i+j) <= n:
        tmp = binin[i:i+j:]
        if words.count(tmp) > 0:
            j += 1
        else:
            words.append(tmp)
            i += j+1
            j = 0
    wobs = len(words)
    pval = 0.5*spc.erfc((mu-wobs)/np.sqrt(2.0*sigma))
    return pval

# test 2.11


def serialtest(binin):
    m = int(log(len(binin), 2) - 3)
    ''' The focus of this test is the frequency of each and every overlapping m-bit pattern across the entire sequence. The purpose of this test is to determine whether the number of occurrences of the 2m m-bit overlapping patterns is approximately the same as would be expected for a random sequence. The pattern can overlap.'''
    n = len(binin)
    hbin = binin+binin[0:m-1:]
    f1a = [hbin[xs:m+xs:] for xs in range(n)]
    oo = set(f1a)
    f1 = [f1a.count(xs)**2 for xs in oo]
    f1 = list(map(f1a.count, oo))
    cou = f1a.count
    f2a = [hbin[xs:m-1+xs:] for xs in range(n)]
    f2 = [f2a.count(xs)**2 for xs in set(f2a)]
    f3a = [hbin[xs:m-2+xs:] for xs in range(n)]
    f3 = [f3a.count(xs)**2 for xs in set(f3a)]
    psim1 = 0
    psim2 = 0
    psim3 = 0
    if m >= 0:
        suss = reduce(su, f1)
        psim1 = 1.0 * 2 ** m * suss / n - n
    if m >= 1:
        suss = reduce(su, f2)
        psim2 = 1.0 * 2 ** (m - 1) * suss / n - n
    if m >= 2:
        suss = reduce(su, f3)
        psim3 = 1.0 * 2 ** (m - 2) * suss / n - n
    d1 = psim1-psim2
    d2 = psim1-2 * psim2 + psim3
    pval1 = spc.gammaincc(2 ** (m - 2), d1 / 2.0)
    pval2 = spc.gammaincc(2 ** (m - 3), d2 / 2.0)
    return [pval1, pval2]


def cumultativesumstest(binin):
    ''' The focus of this test is the maximal excursion (from zero) of the random walk defined by the cumulative sum of adjusted (-1, +1) digits in the sequence. The purpose of the test is to determine whether the cumulative sum of the partial sequences occurring in the tested sequence is too large or too small relative to the expected behavior of that cumulative sum for random sequences.  This cumulative sum may be considered as a random walk. For a random sequence, the random walk should be near zero. For non-random sequences, the excursions of this random walk away from zero will be too large.'''
    n = len(binin)
    ss = [int(el) for el in binin]
    sc = list(map(sumi, ss))
    cs = np.cumsum(sc)
    z = max(abs(cs))
    ra = 0
    start = int(np.floor(0.25 * np.floor(-n / z) + 1))
    stop = int(np.floor(0.25 * np.floor(n / z) - 1))
    pv1 = []
    for k in range(start, stop + 1):
        pv1.append(sst.norm.cdf((4 * k + 1) * z / np.sqrt(n)) -
                   sst.norm.cdf((4 * k - 1) * z / np.sqrt(n)))
    start = int(np.floor(0.25 * np.floor(-n / z - 3)))
    stop = int(np.floor(0.25 * np.floor(n / z) - 1))
    pv2 = []
    for k in range(start, stop + 1):
        pv2.append(sst.norm.cdf((4 * k + 3) * z / np.sqrt(n)) -
                   sst.norm.cdf((4 * k + 1) * z / np.sqrt(n)))
    pval = 1
    pval -= reduce(su, pv1)
    pval += reduce(su, pv2)

    return pval


def cumultativesumstestreverse(binin):
    '''The focus of this test is the maximal excursion (from zero) of the random walk defined by the cumulative sum of adjusted (-1, +1) digits in the sequence. The purpose of the test is to determine whether the cumulative sum of the partial sequences occurring in the tested sequence is too large or too small relative to the expected behavior of that cumulative sum for random sequences.  This cumulative sum may be considered as a random walk. For a random sequence, the random walk should be near zero. For non-random sequences, the excursions of this random walk away from zero will be too large. '''
    pval = cumultativesumstest(binin[::-1])
    return pval


def pik(k, x):
    if k == 0:
        out = 1-1.0/(2*np.abs(x))
    elif k >= 5:
        out = (1.0/(2*np.abs(x)))*(1-1.0/(2*np.abs(x)))**4
    else:
        out = (1.0/(4*x*x))*(1-1.0/(2*np.abs(x)))**(k-1)
    return out


def randomexcursionstest(binin):
    ''' The focus of this test is the number of cycles having exactly K visits in a cumulative sum random walk. The cumulative sum random walk is found if partial sums of the (0,1) sequence are adjusted to (-1, +1). A random excursion of a random walk consists of a sequence of n steps of unit length taken at random that begin at and return to the origin. The purpose of this test is to determine if the number of visits to a state within a random walk exceeds what one would expect for a random sequence.'''
    xvals = [-4, -3, -2, -1, 1, 2, 3, 4]
    ss = [int(el) for el in binin]
    sc = list(map(sumi, ss))
    cumsum = np.cumsum(sc)
    cumsum = np.append(cumsum, 0)
    cumsum = np.append(0, cumsum)
    posi = np.where(cumsum == 0)[0]
    cycles = ([cumsum[posi[x]:posi[x+1]+1] for x in range(len(posi)-1)])
    j = len(cycles)
    sct = []
    for ii in cycles:
        sct.append(([len(np.where(ii == xx)[0]) for xx in xvals]))
    sct = np.transpose(np.clip(sct, 0, 5))
    su = []
    for ii in range(6):
        su.append([(xx == ii).sum() for xx in sct])
    su = np.transpose(su)
    pikt = ([([pik(uu, xx) for uu in range(6)]) for xx in xvals])
    # chitab=1.0*((su-j*pikt)**2)/(j*pikt)
    chitab = np.sum(1.0*(np.array(su)-j*np.array(pikt))
                    ** 2/(j*np.array(pikt)), axis=1)
    pval = ([spc.gammaincc(2.5, cs/2.0) for cs in chitab])
    return pval


def getfreq(linn, nu):
    val = 0
    for (x, y) in linn:
        if x == nu:
            val = y
    return val


def randomexcursionsvarianttest(binin):
    ''' The focus of this test is the number of times that a particular state occurs in a cumulative sum random walk. The purpose of this test is to detect deviations from the expected number of occurrences of various states in the random walk.'''
    ss = [int(el) for el in binin]
    sc = list(map(sumi, ss))
    cs = np.cumsum(sc)
    li = []
    for xs in sorted(set(cs)):
        if np.abs(xs) <= 9:
            li.append([xs, len(np.where(cs == xs)[0])])
    j = getfreq(li, 0) + 1
    pval = []
    for xs in range(-9, 9 + 1):
        if not xs == 0:
            # pval.append([xs, spc.erfc(np.abs(getfreq(li, xs) - j) / np.sqrt(2 * j * (4 * np.abs(xs) - 2)))])
            pval.append(spc.erfc(np.abs(getfreq(li, xs) - j) /
                                 np.sqrt(2 * j * (4 * np.abs(xs) - 2))))
    return pval


def aproximateentropytest(binin, m=5):
    ''' The focus of this test is the frequency of each and every overlapping m-bit pattern. The purpose of the test is to compare the frequency of overlapping blocks of two consecutive/adjacent lengths (m and m+1) against the expected result for a random sequence.'''
    n = len(binin)
    f1a = [(binin + binin[0:m - 1:])[xs:m + xs:] for xs in range(n)]
    f1 = [[xs, f1a.count(xs)] for xs in sorted(set(f1a))]
    f2a = [(binin + binin[0:m:])[xs:m + 1 + xs:] for xs in range(n)]
    f2 = [[xs, f2a.count(xs)] for xs in sorted(set(f2a))]
    c1 = [1.0 * f1[xs][1] / n for xs in range(len(f1))]
    c2 = [1.0 * f2[xs][1] / n for xs in range(len(f2))]
    phi1 = reduce(su, list(map(logo, c1)))
    phi2 = reduce(su, list(map(logo, c2)))
    apen = phi1 - phi2
    chisqr = 2.0 * n * (np.log(2) - apen)
    pval = spc.gammaincc(2 ** (m - 1), chisqr / 2.0)
    return pval


def matrank(mat):  # old function, does not work as advertized - gives the matrix rank, but not binary
    u, s, v = np.linalg.svd(mat)
    rank = np.sum(s > 1e-10)
    return rank


def mrank(matrix):  # matrix rank as defined in the NIST specification
    m = len(matrix)
    leni = len(matrix[0])

    def proc(mat):
        for i in range(m):
            if mat[i][i] == 0:
                for j in range(i+1, m):
                    if mat[j][i] == 1:
                        mat[j], mat[i] = mat[i], mat[j]
                        break
            if mat[i][i] == 1:
                for j in range(i+1, m):
                    if mat[j][i] == 1:
                        mat[j] = [mat[i][x] ^ mat[j][x] for x in range(leni)]
        return mat
    maa = proc(matrix)
    maa.reverse()
    mu = [i[::-1] for i in maa]
    muu = proc(mu)
    ra = np.sum(np.sign([xx.sum() for xx in np.array(mu)]))
    return ra


def binarymatrixranktest(binin, m=32, q=32):
    ''' The focus of the test is the rank of disjoint sub-matrices of the entire sequence. The purpose of this test is to check for linear dependence among fixed length substrings of the original sequence.'''
    p1 = 1.0
    for x in range(1, 50):
        p1 *= 1-(1.0/(2**x))
    p2 = 2*p1
    p3 = 1-p1-p2
    n = len(binin)
    # the input string as numbers, to generate the dot product
    u = [int(el) for el in binin]
    f1a = [u[xs*m:xs*m+m:] for xs in range(floor(n/m))]
    n = len(f1a)
    f2a = [f1a[xs*q:xs*q+q:] for xs in range(floor(n/q))]
    # r=map(matrank,f2a)
    r = list(map(mrank, f2a))
    n = len(r)
    fm = r.count(m)
    fm1 = r.count(m-1)
    chisqr = ((fm-p1*n)**2)/(p1*n)+((fm1-p2*n)**2) / \
        (p2*n)+((n-fm-fm1-p3*n)**2)/(p3*n)
    pval = np.exp(-0.5*chisqr)
    return pval


def lincomplex(binin):
    lenn = len(binin)
    c = b = np.zeros(lenn)
    c[0] = b[0] = 1
    l = 0
    m = -1
    n = 0
    # the input string as numbers, to generate the dot product
    u = [int(el) for el in binin]
    p = 99
    while n < lenn:
        v = u[(n-l):n]  # was n-l..n-1
        v.reverse()
        cc = c[1:l+1]  # was 2..l+1
        d = (u[n]+np.dot(v, cc)) % 2
        if d == 1:
            tmp = c
            p = np.zeros(lenn)
            for i in range(0, l):  # was 1..l+1
                if b[i] == 1:
                    p[i+n-m] = 1
            c = (c+p) % 2
            if l <= 0.5*n:  # was if 2l <= n
                l = n+1-l
                m = n
                b = tmp
        n += 1
    return l

# test 2.10


def linearcomplexitytest(binin, m=500):
    ''' The focus of this test is the length of a generating feedback register. The purpose of this test is to determine whether or not the sequence is complex enough to be considered random. Random sequences are characterized by a longer feedback register. A short feedback register implies non-randomness.'''
    k = 6
    pi = [0.01047, 0.03125, 0.125, 0.5, 0.25, 0.0625, 0.020833]
    avg = 0.5*m + (1.0/36)*(9 + (-1)**(m + 1)) - (m/3.0 + 2.0/9)/2**m
    blocks = stringpart(binin, m)
    bign = len(blocks)
    lc = ([lincomplex(chunk) for chunk in blocks])
    t = ([-1.0*(((-1)**m)*(chunk-avg)+2.0/9) for chunk in lc])
    vg = np.histogram(t, bins=[-9999999999, -2.5, -
                               1.5, -0.5, 0.5, 1.5, 2.5, 9999999999])[0][::-1]
    im = ([((vg[ii]-bign*pi[ii])**2)/(bign*pi[ii]) for ii in range(7)])
    chisqr = reduce(su, im)
    pval = spc.gammaincc(k/2.0, chisqr/2.0)
    return pval


def isRandom(bits):
    result = {}

    def adder(name, p):
        if 'list' in str(type(p)):
            count = 0
            for i in p:
                if 'nan' in str(i):
                    pass
                elif 'e' in str(i):
                    pass
                elif i > 0.01:
                    pass
                else:
                    count += 1
            if count >= (len(p)/2):
                result[name] = False
            else:
                result[name] = True
        elif 'e' in str(p):
            pass
        elif p > 0.01:
            result[name] = True
        else:
            result[name] = False
    try:
        adder('Monobit frequency test', monobitfrequencytest(bits[:100]))
    except:
        pass
    try:
        adder('Block frequency test', blockfrequencytest(bits[:2000]))
    except:
        pass
    try:
        adder('Runs test', runstest(bits))
    except:
        pass
    try:
        adder('Spectral test', spectraltest(bits[:1024]))
    except:
        pass
    try:
        adder('Non-overlapping template matching test',
              nonoverlappingtemplatematchingtest(bits[:1048576], '11111', 8))
    except:
        pass
    try:
        adder('Overlapping template matching test',
              overlappingtemplatematchingtest(bits[:998976], '0000001', 12, 5))
    except:
        pass
    try:
        adder('Serial test', serialtest(bits[:500]))
    except:
        pass
    try:
        adder('Cumultative sums test', cumultativesumstest(bits[:100]))
    except:
        pass
    try:
        adder('Aproximate entropy test', aproximateentropytest(bits[:500], 5))
    except:
        pass
    try:
        adder('Random excursions variant test',
              randomexcursionsvarianttest(bits[:1000000]))
    except:
        pass
    try:
        adder('Linear complexity test',
              linearcomplexitytest(bits[:1000000], 10))
    except:
        pass
    try:
        adder('Longest runs test', longestrunones10000(bits))
    except:
        pass
    try:
        adder('Maurers universal statistic test',
              maurersuniversalstatistictest(bits[:387840], 6, 640))
    except:
        pass
    try:
        adder('Random excursions test', randomexcursionstest(bits[:1000000]))
    except:
        pass
    return result




"""
Base utilities and core functionality for VulnScan
"""

# Import base utilities
from .utils import *
from .sendRequest import *
from .extractHTMLInformation import *
from .rangeFinder import *
from .formEvaluate import *
from .modifyManipulateAPI import *
from .crawl import *
from .prompt import *
from .formParser import *
from .entropy import *
from .colors import *
from .testing import *



import boto3
from google.cloud import storage
from azure.storage.blob import BlobServiceClient
import requests
from urllib.parse import urlparse, urljoin  # Add urljoin here
import json
import time
import socket
import ssl


class CloudSecurityScanner:
    def __init__(self, target):
        self.target = target
        self.results = []
        self.cloud_provider = self.detect_cloud_provider()

    def detect_cloud_provider(self):
        """Detect the cloud provider based on the target"""
        parsed_url = urlparse(self.target)
        domain = parsed_url.netloc

        if 'amazonaws.com' in domain:
            return 'AWS'
        elif 'azurewebsites.net' in domain or 'blob.core.windows.net' in domain:
            return 'Azure'
        elif 'appspot.com' in domain or 'storage.googleapis.com' in domain:
            return 'GCP'
        return None

    def scan_aws_resources(self):
        """Scan AWS resources for security misconfigurations"""
        # Check for S3 bucket exposure
        if 's3.amazonaws.com' in self.target:
            try:
                bucket_name = self.target.split('//')[1].split('.')[0]
                s3 = boto3.client('s3')

                # Check if bucket is publicly accessible
                try:
                    s3.head_bucket(Bucket=bucket_name)
                    self.results.append({
                        'type': 'S3 Bucket Exposure',
                        'severity': 'High',
                        'description': f"S3 bucket {bucket_name} is publicly accessible",
                        'resource': self.target
                    })
                except Exception as e:
                    if 'AccessDenied' in str(e):
                        print(
                            f"S3 bucket {bucket_name} is not publicly accessible")
            except Exception as e:
                print(f"Error checking S3 bucket: {e}")

        # Check for EC2 metadata service exposure
        try:
            metadata_url = "http://169.254.169.254/latest/meta-data/"
            response = requests.get(metadata_url, timeout=2)
            if response.status_code == 200:
                self.results.append({
                    'type': 'EC2 Metadata Exposure',
                    'severity': 'Critical',
                    'description': 'EC2 metadata service is accessible from the target',
                    'resource': metadata_url
                })
        except:
            pass

    def scan_azure_resources(self):
        """Scan Azure resources for security misconfigurations"""
        # Check for Azure blob storage exposure
        if 'blob.core.windows.net' in self.target:
            try:
                blob_service_client = BlobServiceClient.from_blob_url(
                    self.target)
                container_client = blob_service_client.get_container_client()

                # Try to list blobs
                blobs = container_client.list_blobs()
                self.results.append({
                    'type': 'Azure Blob Exposure',
                    'severity': 'High',
                    'description': f"Azure blob storage is publicly accessible: {self.target}",
                    'resource': self.target
                })
            except Exception as e:
                print(f"Azure blob storage is not publicly accessible: {e}")

    def scan_gcp_resources(self):
        """Scan GCP resources for security misconfigurations"""
        # Check for GCP storage bucket exposure
        if 'storage.googleapis.com' in self.target:
            try:
                bucket_name = self.target.split('//')[1].split('/')[0]
                storage_client = storage.Client()
                bucket = storage_client.bucket(bucket_name)

                # Check if bucket is publicly accessible
                if bucket.make_public():
                    self.results.append({
                        'type': 'GCP Storage Exposure',
                        'severity': 'High',
                        'description': f"GCP storage bucket is publicly accessible: {bucket_name}",
                        'resource': self.target
                    })
            except Exception as e:
                print(f"GCP storage bucket is not publicly accessible: {e}")

    def scan_kubernetes_exposure(self):
        """Scan for Kubernetes API server exposure"""
        k8s_paths = [
            '/api/v1/namespaces',
            '/api/v1/pods',
            '/apis/extensions/v1beta1',
            '/apis/apps/v1',
            '/openapi/v2'
        ]

        for path in k8s_paths:
            url = urljoin(self.target, path)
            try:
                response = requests.get(url, timeout=5)
                if response.status_code == 200:
                    self.results.append({
                        'type': 'Kubernetes API Exposure',
                        'severity': 'Critical',
                        'description': f"Kubernetes API endpoint is exposed: {url}",
                        'resource': url
                    })
            except:
                pass

    def run_scan(self):
        """Run all cloud security scans"""
        if self.cloud_provider == 'AWS':
            self.scan_aws_resources()
        elif self.cloud_provider == 'Azure':
            self.scan_azure_resources()
        elif self.cloud_provider == 'GCP':
            self.scan_gcp_resources()

        # Always check for Kubernetes exposure regardless of cloud provider
        self.scan_kubernetes_exposure()

        return self.results



import matplotlib.pyplot as plt
import seaborn as sns
from reportlab.lib.pagesizes import letter
from reportlab.platypus import SimpleDocTemplate, Paragraph, Spacer, Table, TableStyle, Image
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.lib.units import inch
from reportlab.lib import colors
from datetime import datetime
import json
import os
import base64
from io import BytesIO


class AdvancedSecurityReporter:
    def __init__(self, target, scan_results):
        self.target = target
        self.scan_results = scan_results
        self.timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.report_dir = f"reports/{self.target.replace('https://', '').replace('http://', '').replace('/', '_')}"
        os.makedirs(self.report_dir, exist_ok=True)

    def calculate_risk_score(self):
        """Calculate overall risk score based on findings"""
        severity_weights = {
            'Critical': 10,
            'High': 7,
            'Medium': 4,
            'Low': 1
        }

        total_score = 0
        max_score = 0

        # Flatten all findings
        all_findings = self._flatten_findings()

        for finding in all_findings:
            if isinstance(finding, dict):
                severity = finding.get('severity', 'Low')
                total_score += severity_weights.get(severity, 0)
                max_score += severity_weights.get('Critical', 10)

        # Normalize to 0-100 scale
        if max_score > 0:
            risk_score = min(100, int((total_score / max_score) * 100))
        else:
            risk_score = 0

        return risk_score

    def _flatten_findings(self):
        """Flatten nested findings structure"""
        all_findings = []

        for category, findings in self.scan_results.get('findings', {}).items():
            if isinstance(findings, list):
                # Convert simple string results to dict format
                for finding in findings:
                    if isinstance(finding, str):
                        all_findings.append({
                            'type': category.replace('_', ' ').title(),
                            'severity': 'Info',
                            'description': finding
                        })
                    elif isinstance(finding, dict):
                        all_findings.append(finding)
            elif isinstance(findings, dict):
                # Handle nested structures like api_vulnerabilities
                for subcategory, subfindings in findings.items():
                    if isinstance(subfindings, list):
                        all_findings.extend(subfindings)
                    elif isinstance(subfindings, dict):
                        all_findings.append(subfindings)
            elif findings:
                # Handle single results
                all_findings.append({
                    'type': category.replace('_', ' ').title(),
                    'severity': 'Info',
                    'description': str(findings)
                })

        return all_findings

    def get_risk_color(self, risk_score):
        """Get color based on risk score"""
        if risk_score >= 90:
            return colors.red
        elif risk_score >= 70:
            return colors.orange
        elif risk_score >= 40:
            return colors.yellow
        else:
            return colors.green

    def generate_visualizations(self):
        """Generate visualizations for the report"""
        try:
            # Count vulnerabilities by severity
            severity_counts = {'Critical': 0, 'High': 0, 'Medium': 0, 'Low': 0}

            all_findings = self._flatten_findings()

            for finding in all_findings:
                if isinstance(finding, dict):
                    severity = finding.get('severity', 'Low')
                    if severity in severity_counts:
                        severity_counts[severity] += 1

            # Create bar chart
            plt.figure(figsize=(10, 6))
            sns.barplot(x=list(severity_counts.keys()),
                        y=list(severity_counts.values()))
            plt.title('Vulnerability Severity Distribution')
            plt.xlabel('Severity')
            plt.ylabel('Count')
            plt.tight_layout()
            plt.savefig(f"{self.report_dir}/severity_distribution.png")
            plt.close()

            # Create pie chart
            plt.figure(figsize=(8, 8))
            plt.pie(list(severity_counts.values()), labels=list(
                severity_counts.keys()), autopct='%1.1f%%')
            plt.title('Vulnerability Severity Distribution')
            plt.tight_layout()
            plt.savefig(f"{self.report_dir}/severity_pie.png")
            plt.close()

            # Create risk gauge
            risk_score = self.calculate_risk_score()
            fig, ax = plt.subplots(figsize=(8, 6))

            # Create gauge
            gauge_colors = ['green', 'yellow', 'orange', 'red']
            gauge_limits = [0, 25, 50, 75, 100]

            # Draw gauge
            for i in range(len(gauge_limits)-1):
                ax.barh(0, gauge_limits[i+1]-gauge_limits[i], left=gauge_limits[i],
                        color=gauge_colors[i], height=0.5)

            # Draw needle
            ax.barh(0, risk_score, color='black', height=0.1)

            # Set labels
            ax.set_xlim(0, 100)
            ax.set_ylim(-0.5, 0.5)
            ax.set_yticks([])
            ax.set_xticks([0, 25, 50, 75, 100])
            ax.set_title(f'Risk Score: {risk_score}/100')

            plt.tight_layout()
            plt.savefig(f"{self.report_dir}/risk_gauge.png")
            plt.close()

        except Exception as e:
            print(f"Error generating visualizations: {e}")
            # Create placeholder images if visualization fails
            for img_name in ['severity_distribution.png', 'severity_pie.png', 'risk_gauge.png']:
                try:
                    fig, ax = plt.subplots(figsize=(6, 4))
                    ax.text(0.5, 0.5, 'Visualization Error',
                            ha='center', va='center')
                    ax.axis('off')
                    plt.savefig(f"{self.report_dir}/{img_name}")
                    plt.close()
                except:
                    pass

    def generate_pdf_report(self):
        """Generate comprehensive PDF report"""
        doc = SimpleDocTemplate(
            f"{self.report_dir}/security_report.pdf", pagesize=letter)
        styles = getSampleStyleSheet()
        story = []

        # Title
        title_style = ParagraphStyle(
            'CustomTitle',
            parent=styles['Heading1'],
            fontSize=24,
            spaceAfter=30,
            textColor=colors.darkblue
        )
        story.append(
            Paragraph("Advanced Security Vulnerability Report", title_style))
        story.append(Spacer(1, 12))

        # Target and timestamp
        story.append(
            Paragraph(f"<b>Target:</b> {self.target}", styles['Normal']))
        story.append(
            Paragraph(f"<b>Scan Date:</b> {self.timestamp}", styles['Normal']))
        story.append(Spacer(1, 12))

        # Executive Summary
        story.append(Paragraph("Executive Summary", styles['Heading2']))
        risk_score = self.calculate_risk_score()
        risk_color = self.get_risk_color(risk_score)

        # Fixed color formatting - use hex color without any modification
        story.append(Paragraph(
            f"<b>Risk Score:</b> <font color='{risk_color.hexval()}'>{risk_score}/100</font>", styles['Normal']))
        story.append(Spacer(1, 12))

        # Add visualizations
        self.generate_visualizations()

        # Add risk gauge
        if os.path.exists(f"{self.report_dir}/risk_gauge.png"):
            story.append(
                Image(f"{self.report_dir}/risk_gauge.png", width=4*inch, height=3*inch))
            story.append(Spacer(1, 12))

        # Add severity distribution
        if os.path.exists(f"{self.report_dir}/severity_distribution.png"):
            story.append(Image(
                f"{self.report_dir}/severity_distribution.png", width=6*inch, height=4*inch))
            story.append(Spacer(1, 12))

        # Detailed Findings
        story.append(Paragraph("Detailed Findings", styles['Heading2']))

        # Process all findings
        all_findings = self._flatten_findings()

        if all_findings:
            # Create table for findings
            data = [["Finding Type", "Severity", "Description"]]
            for finding in all_findings:
                if isinstance(finding, dict):
                    finding_type = finding.get(
                        'type', finding.get('vulnerability', 'Unknown'))
                    severity = finding.get('severity', 'Low')
                    description = finding.get('description', 'No description')

                    # Truncate long descriptions
                    if len(description) > 100:
                        description = description[:97] + "..."

                    data.append([finding_type, severity, description])

            table = Table(data)
            table.setStyle(TableStyle([
                ('BACKGROUND', (0, 0), (-1, 0), colors.grey),
                ('TEXTCOLOR', (0, 0), (-1, 0), colors.whitesmoke),
                ('ALIGN', (0, 0), (-1, -1), 'LEFT'),
                ('FONTNAME', (0, 0), (-1, 0), 'Helvetica-Bold'),
                ('FONTSIZE', (0, 0), (-1, 0), 12),
                ('BOTTOMPADDING', (0, 0), (-1, 0), 12),
                ('BACKGROUND', (0, 1), (-1, -1), colors.beige),
                ('GRID', (0, 0), (-1, -1), 1, colors.black),
                ('VALIGN', (0, 0), (-1, -1), 'TOP')
            ]))
            story.append(table)
            story.append(Spacer(1, 12))
        else:
            story.append(
                Paragraph("No vulnerabilities found.", styles['Normal']))
            story.append(Spacer(1, 12))

        # Module-specific findings
        story.append(Paragraph("Module-Specific Findings", styles['Heading2']))

        for category, findings in self.scan_results.items():
            if category in ['target', 'timestamp']:
                continue

            if isinstance(findings, dict) and findings:
                story.append(Paragraph(category.replace(
                    '_', ' ').title(), styles['Heading3']))

                # Handle nested structures
                for subcategory, subfindings in findings.items():
                    if isinstance(subfindings, list) and subfindings:
                        story.append(Paragraph(subcategory.replace(
                            '_', ' ').title(), styles['Heading4']))

                        for finding in subfindings:
                            if isinstance(finding, dict):
                                story.append(Paragraph(
                                    f"<b>Type:</b> {finding.get('type', 'Unknown')}<br/>"
                                    f"<b>Severity:</b> {finding.get('severity', 'Low')}<br/>"
                                    f"<b>Description:</b> {finding.get('description', 'No description')}",
                                    styles['Normal']))
                                story.append(Spacer(1, 6))

            elif isinstance(findings, list) and findings:
                story.append(Paragraph(category.replace(
                    '_', ' ').title(), styles['Heading3']))

                for finding in findings:
                    if isinstance(finding, dict):
                        story.append(Paragraph(
                            f"<b>Type:</b> {finding.get('type', 'Unknown')}<br/>"
                            f"<b>Severity:</b> {finding.get('severity', 'Low')}<br/>"
                            f"<b>Description:</b> {finding.get('description', 'No description')}",
                            styles['Normal']))
                        story.append(Spacer(1, 6))

        # Recommendations
        story.append(Paragraph("Recommendations", styles['Heading2']))
        recommendations = self.generate_recommendations()
        for rec in recommendations:
            story.append(Paragraph(f"• {rec}", styles['Normal']))

        # Build PDF
        doc.build(story)
        return f"{self.report_dir}/security_report.pdf"

    def generate_html_report(self):
        """Generate interactive HTML report"""
        all_findings = self._flatten_findings()
        risk_score = self.calculate_risk_score()

        html_content = f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>Security Report - {self.target}</title>
            <style>
                body {{ font-family: Arial, sans-serif; margin: 20px; }}
                .header {{ background-color: #f0f0f0; padding: 20px; border-radius: 5px; }}
                .section {{ margin-top: 20px; }}
                .finding {{ background-color: #fff; border: 1px solid #ddd; padding: 15px; margin-bottom: 10px; border-radius: 5px; }}
                .critical {{ border-left: 5px solid #d9534f; }}
                .high {{ border-left: 5px solid #f0ad4e; }}
                .medium {{ border-left: 5px solid #5bc0de; }}
                .low {{ border-left: 5px solid #5cb85c; }}
                .risk-score {{ font-size: 48px; font-weight: bold; text-align: center; }}
                .risk-low {{ color: #5cb85c; }}
                .risk-medium {{ color: #f0ad4e; }}
                .risk-high {{ color: #d9534f; }}
                .risk-critical {{ color: #8b0000; }}
                table {{ border-collapse: collapse; width: 100%; }}
                th, td {{ border: 1px solid #ddd; padding: 8px; text-align: left; }}
                th {{ background-color: #f2f2f2; }}
                .module-section {{ margin-top: 30px; border-top: 1px solid #ddd; padding-top: 20px; }}
            </style>
        </head>
        <body>
            <div class="header">
                <h1>Security Vulnerability Report</h1>
                <p><b>Target:</b> {self.target}</p>
                <p><b>Scan Date:</b> {self.timestamp}</p>
            </div>
            
            <div class="section">
                <h2>Executive Summary</h2>
                <div class="risk-score risk-{self.get_risk_class(risk_score)}">{risk_score}/100</div>
            </div>
            
            <div class="section">
                <h2>Findings Summary</h2>
                <table>
                    <tr>
                        <th>Category</th>
                        <th>Critical</th>
                        <th>High</th>
                        <th>Medium</th>
                        <th>Low</th>
                    </tr>
        """

        # Add summary table
        severity_counts = {'Critical': 0, 'High': 0, 'Medium': 0, 'Low': 0}
        for finding in all_findings:
            if isinstance(finding, dict):
                severity = finding.get('severity', 'Low')
                if severity in severity_counts:
                    severity_counts[severity] += 1

        html_content += f"""
                    <tr>
                        <td>All Findings</td>
                        <td>{severity_counts['Critical']}</td>
                        <td>{severity_counts['High']}</td>
                        <td>{severity_counts['Medium']}</td>
                        <td>{severity_counts['Low']}</td>
                    </tr>
                </table>
            </div>
            
            <div class="section">
                <h2>Detailed Findings</h2>
        """

        # Add detailed findings
        for finding in all_findings:
            if isinstance(finding, dict):
                severity = finding.get('severity', 'Low').lower()
                finding_type = finding.get(
                    'type', finding.get('vulnerability', 'Unknown'))
                description = finding.get('description', 'No description')

                html_content += f"""
                <div class="finding {severity}">
                    <h4>{finding_type}</h4>
                    <p><b>Severity:</b> {finding.get('severity', 'Low')}</p>
                    <p><b>Description:</b> {description}</p>
                </div>
                """

        # Add module-specific findings
        html_content += """
            <div class="section">
                <h2>Module-Specific Findings</h2>
        """

        for category, findings in self.scan_results.items():
            if category in ['target', 'timestamp']:
                continue

            if isinstance(findings, dict) and findings:
                html_content += f"""
                <div class="module-section">
                    <h3>{category.replace('_', ' ').title()}</h3>
                """

                for subcategory, subfindings in findings.items():
                    if isinstance(subfindings, list) and subfindings:
                        html_content += f"""
                        <h4>{subcategory.replace('_', ' ').title()}</h4>
                        """

                        for finding in subfindings:
                            if isinstance(finding, dict):
                                severity = finding.get(
                                    'severity', 'Low').lower()
                                finding_type = finding.get('type', 'Unknown')
                                description = finding.get(
                                    'description', 'No description')

                                html_content += f"""
                                <div class="finding {severity}">
                                    <h5>{finding_type}</h5>
                                    <p><b>Severity:</b> {finding.get('severity', 'Low')}</p>
                                    <p><b>Description:</b> {description}</p>
                                </div>
                                """

                html_content += "</div>"

        # Add recommendations
        html_content += """
            </div>
            
            <div class="section">
                <h2>Recommendations</h2>
                <ul>
        """

        for rec in self.generate_recommendations():
            html_content += f"<li>{rec}</li>"

        html_content += """
                </ul>
            </div>
        </body>
        </html>
        """

        with open(f"{self.report_dir}/security_report.html", 'w') as f:
            f.write(html_content)

        return f"{self.report_dir}/security_report.html"

    def get_risk_class(self, risk_score):
        """Get CSS class for risk score"""
        if risk_score >= 90:
            return 'critical'
        elif risk_score >= 70:
            return 'high'
        elif risk_score >= 40:
            return 'medium'
        else:
            return 'low'

    def generate_json_report(self):
        """Generate JSON report for integration with other tools"""
        report = {
            "target": self.target,
            "scan_date": self.timestamp,
            "risk_score": self.calculate_risk_score(),
            "findings": self.scan_results,
            "recommendations": self.generate_recommendations()
        }

        with open(f"{self.report_dir}/security_report.json", 'w') as f:
            json.dump(report, f, indent=4)

        return f"{self.report_dir}/security_report.json"

    def generate_recommendations(self):
        """Generate recommendations based on findings"""
        recommendations = []
        all_findings = self._flatten_findings()

        # Check for critical vulnerabilities
        for finding in all_findings:
            if isinstance(finding, dict) and finding.get('severity') == 'Critical':
                vuln_type = finding.get('type', 'Unknown')
                if 'SQL Injection' in vuln_type:
                    recommendations.append(
                        "Implement parameterized queries or prepared statements to prevent SQL injection attacks.")
                elif 'XSS' in vuln_type:
                    recommendations.append(
                        "Implement Content Security Policy (CSP) and sanitize user input to prevent XSS attacks.")
                elif 'Subdomain Takeover' in vuln_type:
                    recommendations.append(
                        "Remove unused DNS records or verify ownership of all subdomains.")
                elif 'JWT' in vuln_type:
                    recommendations.append(
                        "Use strong secrets for JWT tokens and avoid using the 'none' algorithm.")

        # General recommendations
        recommendations.append(
            "Implement a Web Application Firewall (WAF) to protect against common attacks.")
        recommendations.append(
            "Regularly update and patch all software and dependencies.")
        recommendations.append(
            "Conduct regular security assessments and penetration testing.")
        recommendations.append(
            "Implement proper logging and monitoring for security events.")
        recommendations.append(
            "Follow the principle of least privilege for all user accounts and services.")

        return recommendations

    def generate_all_reports(self):
        """Generate all report formats"""
        pdf_report = self.generate_pdf_report()
        json_report = self.generate_json_report()
        html_report = self.generate_html_report()

        return {
            'pdf': pdf_report,
            'json': json_report,
            'html': html_report
        }



import requests
from urllib.parse import urljoin
import json
import re
from bs4 import BeautifulSoup
import time


class GraphQLSecurityTester:
    def __init__(self, target_url):
        self.target_url = target_url
        self.headers = {'Content-Type': 'application/json'}
        self.results = []

    def test_graphql_introspection(self):
        """Test for GraphQL introspection exposure"""
        introspection_query = {
            "query": """
            query IntrospectionQuery {
                __schema {
                    queryType { name }
                    mutationType { name }
                    subscriptionType { name }
                    types {
                        ...FullType
                    }
                    directives {
                        name
                        description
                        locations
                        args {
                            ...InputValue
                        }
                    }
                }
            }
            fragment FullType on __Type {
                kind
                name
                description
                fields(includeDeprecated: true) {
                    name
                    description
                    args {
                        ...InputValue
                    }
                    type {
                        ...TypeRef
                    }
                    isDeprecated
                    deprecationReason
                }
                inputFields {
                    ...InputValue
                }
                interfaces {
                    ...TypeRef
                }
                enumValues(includeDeprecated: true) {
                    name
                    description
                    isDeprecated
                    deprecationReason
                }
                possibleTypes {
                    ...TypeRef
                }
            }
            fragment InputValue on __InputValue {
                name
                description
                type { ...TypeRef }
                defaultValue
            }
            fragment TypeRef on __Type {
                kind
                name
                ofType {
                    kind
                    name
                    ofType {
                        kind
                        name
                        ofType {
                            kind
                            name
                            ofType {
                                kind
                                name
                                ofType {
                                    kind
                                    name
                                    ofType {
                                        kind
                                        name
                                    }
                                }
                            }
                        }
                    }
                }
            }
            """
        }

        try:
            response = requests.post(
                urljoin(self.target_url, '/graphql'),
                json=introspection_query,
                headers=self.headers
            )

            if response.status_code == 200 and '__schema' in response.text:
                self.results.append({
                    'type': 'GraphQL Introspection',
                    'severity': 'High',
                    'description': 'GraphQL introspection is enabled, exposing schema details',
                    'endpoint': urljoin(self.target_url, '/graphql')
                })

        except Exception as e:
            print(f"Error testing GraphQL introspection: {e}")

    def test_graphql_dos(self):
        """Test for GraphQL DoS vulnerabilities"""
        dos_query = {
            "query": """
            query {
                posts {
                    id
                    title
                    author {
                        id
                        name
                        posts {
                            id
                            title
                            author {
                                id
                                name
                                posts {
                                    id
                                    title
                                    author {
                                        id
                                        name
                                    }
                                }
                            }
                        }
                    }
                }
            }
            """
        }

        try:
            start_time = time.time()
            response = requests.post(
                urljoin(self.target_url, '/graphql'),
                json=dos_query,
                headers=self.headers,
                timeout=10
            )
            response_time = time.time() - start_time

            if response_time > 5:
                self.results.append({
                    'type': 'GraphQL DoS Vulnerability',
                    'severity': 'High',
                    'description': f"GraphQL endpoint is vulnerable to DoS attacks (response time: {response_time:.2f}s)",
                    'endpoint': urljoin(self.target_url, '/graphql')
                })

        except Exception as e:
            print(f"Error testing GraphQL DoS: {e}")

    def run_tests(self):
        """Run all GraphQL security tests"""
        self.test_graphql_introspection()
        self.test_graphql_dos()
        return self.results


class APISecurityTester:
    def __init__(self, target_url):
        self.target_url = target_url
        self.headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36'
        }
        self.results = []

    def discover_api_endpoints(self):
        """Discover API endpoints in the target"""
        # Common API paths
        api_paths = [
            '/api', '/rest', '/graphql', '/v1', '/v2', '/v3',
            '/swagger', '/openapi.json', '/api-docs', '/docs'
        ]

        discovered_endpoints = []

        for path in api_paths:
            url = urljoin(self.target_url, path)
            try:
                response = requests.get(url, headers=self.headers, timeout=5)
                if response.status_code == 200:
                    discovered_endpoints.append({
                        'url': url,
                        'status_code': response.status_code,
                        'content_type': response.headers.get('Content-Type', '')
                    })

                    # Check for specific API indicators
                    if 'swagger' in response.text.lower() or 'openapi' in response.text.lower():
                        self.results.append({
                            'type': 'Exposed API Documentation',
                            'severity': 'Medium',
                            'description': f"API documentation is exposed at {url}",
                            'endpoint': url
                        })

                    if 'graphql' in response.text.lower():
                        self.results.append({
                            'type': 'GraphQL Endpoint',
                            'severity': 'Medium',
                            'description': f"GraphQL endpoint discovered at {url}",
                            'endpoint': url
                        })
            except:
                pass

        return discovered_endpoints

    def test_rest_api_security(self, endpoint):
        """Test REST API endpoints for common vulnerabilities"""
        # Test for common REST API vulnerabilities
        test_cases = [
            # Test for IDOR
            {'method': 'GET', 'params': {'id': 1}},
            # Test for SQL injection
            {'method': 'GET', 'params': {'id': "1' OR '1'='1"}},
            # Test for XSS
            {'method': 'GET', 'params': {
                'search': "<script>alert('XSS')</script>"}},
        ]

        for test in test_cases:
            try:
                if test['method'] == 'GET':
                    response = requests.get(
                        endpoint['url'],
                        params=test['params'],
                        headers=self.headers,
                        timeout=5
                    )

                # Check for vulnerability indicators
                if 'SQL syntax' in response.text.lower():
                    self.results.append({
                        'type': 'SQL Injection',
                        'severity': 'High',
                        'description': f"Potential SQL injection in API endpoint: {endpoint['url']}",
                        'endpoint': endpoint['url']
                    })

                if '<script>' in response.text and 'XSS' in response.text:
                    self.results.append({
                        'type': 'XSS',
                        'severity': 'High',
                        'description': f"Potential XSS in API endpoint: {endpoint['url']}",
                        'endpoint': endpoint['url']
                    })

            except:
                pass

    def test_authentication_bypass(self):
        """Test for authentication bypass vulnerabilities"""
        # Common authentication endpoints
        auth_paths = ['/login', '/auth', '/signin', '/api/login', '/api/auth']

        for path in auth_paths:
            url = urljoin(self.target_url, path)
            try:
                # Test without authentication
                response = requests.get(url, headers=self.headers, timeout=5)

                # Test with common bypass techniques
                bypass_headers = self.headers.copy()
                bypass_headers['X-Original-URL'] = '/admin'
                bypass_headers['X-Rewrite-URL'] = '/admin'

                bypass_response = requests.get(
                    url, headers=bypass_headers, timeout=5)

                # Check if bypass was successful
                if bypass_response.status_code == 200 and 'admin' in bypass_response.text.lower():
                    self.results.append({
                        'type': 'Authentication Bypass',
                        'severity': 'Critical',
                        'description': f"Authentication bypass possible at {url}",
                        'endpoint': url
                    })

            except:
                pass

    def test_rate_limiting(self):
        """Test for rate limiting vulnerabilities"""
        # Test with rapid requests
        test_endpoint = urljoin(self.target_url, '/api')

        try:
            for i in range(10):
                response = requests.get(
                    test_endpoint, headers=self.headers, timeout=5)
                if response.status_code == 429:
                    # Rate limiting is in place
                    return

            # If we get here, rate limiting might not be in place
            self.results.append({
                'type': 'Missing Rate Limiting',
                'severity': 'Medium',
                'description': f"No rate limiting detected at {test_endpoint}",
                'endpoint': test_endpoint
            })

        except:
            pass

    def run_tests(self):
        """Run all API security tests"""
        print(f"[*] Running API security tests for {self.target_url}")

        # Discover API endpoints
        endpoints = self.discover_api_endpoints()

        # Test discovered endpoints
        for endpoint in endpoints:
            self.test_rest_api_security(endpoint)

        # Test for authentication bypass
        self.test_authentication_bypass()

        # Test rate limiting
        self.test_rate_limiting()

        return self.results




"""
Security scanning modules for VulnScan
"""

# Import all modules
from .advanced_api_checking import *
from .advanced_reporting import *
from .ai_powered_checking import *
from .cloud_vulnerability_checking import *
from .domain_passive_active_check import *
from .modern_security_platform import *
from .web_app_checking import *
from .security_tool_integration import *



import requests
from bs4 import BeautifulSoup
import re
import json
from urllib.parse import urljoin, urlparse
import time
import random
import string
import concurrent.futures


class AdvancedWebAppTester:
    def __init__(self, target_url):
        self.target_url = target_url
        self.session = requests.Session()
        self.session.headers.update({
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36'
        })
        self.results = []
        self.crawled_urls = set()

    def crawl_website(self, max_pages=50):
        """Crawl the website to discover pages and forms"""
        def crawl_page(url, depth=0):
            if depth > 2 or url in self.crawled_urls or len(self.crawled_urls) >= max_pages:
                return

            self.crawled_urls.add(url)

            try:
                response = self.session.get(url, timeout=10)
                if response.status_code != 200:
                    return

                soup = BeautifulSoup(response.text, 'html.parser')

                # Extract forms
                forms = soup.find_all('form')
                for form in forms:
                    self.analyze_form(form, url)

                # Extract links
                for link in soup.find_all('a', href=True):
                    href = link.get('href')
                    if href and not href.startswith('javascript:'):
                        next_url = urljoin(url, href)
                        if self.is_same_domain(next_url):
                            crawl_page(next_url, depth + 1)
            except Exception as e:
                print(f"Error crawling {url}: {e}")

        crawl_page(self.target_url)
        return self.crawled_urls

    def is_same_domain(self, url):
        """Check if URL belongs to the same domain"""
        target_domain = urlparse(self.target_url).netloc
        url_domain = urlparse(url).netloc
        return target_domain == url_domain

    def analyze_form(self, form, page_url):
        """Analyze a form for security vulnerabilities"""
        action = form.get('action', '')
        method = form.get('method', 'get').lower()
        inputs = form.find_all('input')

        # Construct form URL
        form_url = urljoin(page_url, action)

        # Test for CSRF
        csrf_result = self.test_csrf_protection(form, form_url)
        if csrf_result:
            self.results.append(csrf_result)

        # Test for SQL injection
        sqli_result = self.test_sql_injection(form, form_url)
        if sqli_result:
            self.results.append(sqli_result)

        # Test for XSS
        xss_result = self.test_xss(form, form_url)
        if xss_result:
            self.results.append(xss_result)

        # Test for parameter tampering
        tampering_result = self.test_parameter_tampering(form, form_url)
        if tampering_result:
            self.results.append(tampering_result)

    def test_csrf_protection(self, form, form_url):
        """Test form for CSRF protection"""
        # Check for CSRF tokens
        csrf_token_names = ['csrf_token', 'authenticity_token',
                            '_token', 'anticsrf', 'csrfmiddlewaretoken']

        has_csrf_token = False
        for input_tag in form.find_all('input'):
            name = input_tag.get('name', '').lower()
            if any(token_name in name for token_name in csrf_token_names):
                has_csrf_token = True
                break

        if not has_csrf_token:
            return {
                'type': 'Missing CSRF Protection',
                'severity': 'High',
                'description': f"Form at {form_url} does not have CSRF protection",
                'url': form_url
            }

        return None

    def test_sql_injection(self, form, form_url):
        """Test form for SQL injection vulnerabilities"""
        method = form.get('method', 'get').lower()

        # SQL injection payloads
        sqli_payloads = [
            "' OR '1'='1",
            "' OR 1=1--",
            "' UNION SELECT NULL--",
            "' AND (SELECT COUNT(*) FROM information_schema.tables)>0--",
            "' AND SLEEP(5)--",
            "'; DROP TABLE users--"
        ]

        for payload in sqli_payloads:
            data = {}

            # Prepare form data
            for input_tag in form.find_all('input'):
                name = input_tag.get('name')
                input_type = input_tag.get('type', 'text')

                if name:
                    if input_type in ['text', 'search', 'hidden']:
                        data[name] = payload
                    elif input_type == 'submit':
                        data[name] = input_tag.get('value', 'Submit')

            # Send request
            try:
                start_time = time.time()
                if method == 'post':
                    response = self.session.post(
                        form_url, data=data, timeout=10)
                else:
                    response = self.session.get(
                        form_url, params=data, timeout=10)
                response_time = time.time() - start_time

                # Check for SQL injection indicators
                if any(indicator in response.text.lower() for indicator in ['sql syntax', 'mysql_fetch', 'ora-00936', 'microsoft ole db provider']):
                    return {
                        'type': 'SQL Injection',
                        'severity': 'Critical',
                        'description': f"SQL injection vulnerability detected at {form_url} with payload: {payload}",
                        'url': form_url,
                        'payload': payload
                    }

                # Check for time-based SQL injection
                if response_time > 5:
                    return {
                        'type': 'Time-Based SQL Injection',
                        'severity': 'High',
                        'description': f"Time-based SQL injection detected at {form_url} with payload: {payload}",
                        'url': form_url,
                        'payload': payload
                    }
            except Exception as e:
                print(f"Error testing SQL injection: {e}")

        return None

    def test_xss(self, form, form_url):
        """Test form for XSS vulnerabilities"""
        method = form.get('method', 'get').lower()

        # XSS payloads
        xss_payloads = [
            "<script>alert('XSS')</script>",
            "<img src='x' onerror=\"alert('XSS')\">",
            "<svg onload=alert('XSS')>",
            "javascript:alert('XSS')",
            "<iframe src=javascript:alert('XSS')>",
            "'\"><script>alert('XSS')</script>"
        ]

        for payload in xss_payloads:
            data = {}

            # Prepare form data
            for input_tag in form.find_all('input'):
                name = input_tag.get('name')
                input_type = input_tag.get('type', 'text')

                if name:
                    if input_type in ['text', 'search', 'hidden']:
                        data[name] = payload
                    elif input_type == 'submit':
                        data[name] = input_tag.get('value', 'Submit')

            # Send request
            try:
                if method == 'post':
                    response = self.session.post(
                        form_url, data=data, timeout=10)
                else:
                    response = self.session.get(
                        form_url, params=data, timeout=10)

                # Check for XSS in response
                if payload in response.text:
                    return {
                        'type': 'XSS',
                        'severity': 'High',
                        'description': f"XSS vulnerability detected at {form_url} with payload: {payload}",
                        'url': form_url,
                        'payload': payload
                    }
            except Exception as e:
                print(f"Error testing XSS: {e}")

        return None

    def test_parameter_tampering(self, form, form_url):
        """Test form for parameter tampering vulnerabilities"""
        method = form.get('method', 'get').lower()

        # Get original form data
        original_data = {}
        for input_tag in form.find_all('input'):
            name = input_tag.get('name')
            value = input_tag.get('value', '')

            if name:
                original_data[name] = value

        # Try tampering with parameters
        for param in original_data:
            if param.lower() in ['id', 'user_id', 'product_id', 'order_id']:
                tampered_data = original_data.copy()
                tampered_data[param] = '1'  # Try to access resource with ID 1

                try:
                    if method == 'post':
                        response = self.session.post(
                            form_url, data=tampered_data, timeout=10)
                    else:
                        response = self.session.get(
                            form_url, params=tampered_data, timeout=10)

                    # Check if we got access to a different resource
                    if 'unauthorized' not in response.text.lower() and 'access denied' not in response.text.lower():
                        return {
                            'type': 'Parameter Tampering',
                            'severity': 'Medium',
                            'description': f"Parameter tampering vulnerability detected at {form_url} for parameter: {param}",
                            'url': form_url,
                            'parameter': param
                        }
                except Exception as e:
                    print(f"Error testing parameter tampering: {e}")

        return None

    def test_jwt_security(self):
        """Test for JWT security vulnerabilities"""
        # Look for JWT tokens in responses
        response = self.session.get(self.target_url, timeout=10)

        # JWT pattern
        jwt_pattern = r'eyJ[A-Za-z0-9_-]*\.[A-Za-z0-9_-]*\.[A-Za-z0-9_-]*'
        tokens = re.findall(jwt_pattern, response.text)

        for token in tokens:
            # Test for 'none' algorithm vulnerability
            try:
                parts = token.split('.')
                header = json.loads(base64.urlsafe_b64decode(parts[0] + '=='))

                if 'alg' in header and header['alg'].lower() == 'none':
                    self.results.append({
                        'type': 'JWT None Algorithm',
                        'severity': 'High',
                        'description': f"JWT token uses 'none' algorithm: {token[:20]}...",
                        'url': self.target_url
                    })

                # Test for weak secret
                if 'alg' in header and header['alg'].lower() == 'hs256':
                    # Try to crack the secret with common passwords
                    common_secrets = ['secret', 'password',
                                      'api_key', 'jwt_secret']
                    for secret in common_secrets:
                        try:
                            decoded = jwt.decode(
                                token, secret, algorithms=['HS256'])
                            self.results.append({
                                'type': 'Weak JWT Secret',
                                'severity': 'High',
                                'description': f"JWT token cracked with weak secret: {secret}",
                                'url': self.target_url
                            })
                            break
                        except:
                            pass
            except:
                pass

    def test_api_security(self):
        """Test API endpoints for security vulnerabilities"""
        # Common API paths
        api_paths = ['/api', '/rest', '/graphql',
                     '/v1', '/v2', '/swagger', '/openapi.json']

        for path in api_paths:
            url = urljoin(self.target_url, path)
            try:
                response = self.session.get(url, timeout=10)
                if response.status_code == 200:
                    # Test for API-specific vulnerabilities
                    if 'graphql' in path:
                        self.test_graphql_security(url)
                    elif 'swagger' in path or 'openapi' in path:
                        self.results.append({
                            'type': 'Exposed API Documentation',
                            'severity': 'Medium',
                            'description': f"API documentation is exposed at {url}",
                            'url': url
                        })
            except:
                pass

    def test_graphql_security(self, url):
        """Test GraphQL endpoint for security vulnerabilities"""
        # Test for GraphQL introspection
        introspection_query = {
            "query": "{__schema{types{name}}}"
        }

        try:
            response = self.session.post(
                url, json=introspection_query, timeout=10)
            if response.status_code == 200 and '__schema' in response.text:
                self.results.append({
                    'type': 'GraphQL Introspection',
                    'severity': 'High',
                    'description': f"GraphQL introspection is enabled at {url}",
                    'url': url
                })
        except:
            pass

    def run_tests(self):
        """Run all web application security tests"""
        print(f"[*] Starting advanced web app scan for {self.target_url} (UA: {self.session.headers['User-Agent'][:30]}...)")

        # Test XSS
        print("[*] Testing for XSS...")
        self.test_xss_parallel()

        # Test SQL Injection
        print("[*] Testing for SQL Injection...")
        self.test_sqli_parallel()

        # Test JWT security
        print("[*] Testing JWT security...")
        self.test_jwt_security()

        # Test API security
        print("[*] Testing API security...")
        self.test_api_security()

        # Test Security Headers
        print("[*] Testing security headers...")
        self.test_security_headers()

        # Test Cookie Security
        print("[*] Testing cookie security...")
        self.test_cookie_security()

        return self.results

    def test_xss_parallel(self):
        """Test for XSS vulnerabilities using parallelism"""
        try:
            response = self.session.get(self.target_url, timeout=10)
            soup = BeautifulSoup(response.text, 'html.parser')
            forms = soup.find_all('form')
            
            print(f"[*] Found {len(forms)} forms to test for XSS.")

            with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
                futures = [executor.submit(self._test_form_xss, form, response.url) for form in forms]
                concurrent.futures.wait(futures)
                
            # Also test URL parameters
            self._test_url_xss()

        except Exception as e:
            print(f"[-] Error testing XSS: {e}")

    def _test_form_xss(self, form, url):
        """Helper to test a single form for XSS"""
        try:
            action = form.get('action')
            method = form.get('method', 'get').lower()
            inputs = form.find_all('input')
            
            target_url = urljoin(url, action)
            
            for payload in self.xss_payloads:
                data = {}
                for input_tag in inputs:
                    if input_tag.get('type') in ['text', 'search', 'url', 'email', 'password']:
                        data[input_tag.get('name')] = payload
                    else:
                         data[input_tag.get('name')] = input_tag.get('value', '')
                
                try:
                    if method == 'post':
                        res = self.session.post(target_url, data=data, timeout=5)
                    else:
                        res = self.session.get(target_url, params=data, timeout=5)
                    
                    if payload in res.text:
                         self.results.append({
                            'type': 'Reflected XSS',
                            'severity': 'High',
                            'description': f"XSS vulnerability found in form at {target_url} with payload {payload}",
                            'url': target_url,
                            'payload': payload
                        })
                         return # Stop after finding one XSS in this form to save time
                except:
                    pass
        except Exception as e:
            pass

    def test_sqli_parallel(self):
        """Test for SQL Injection vulnerabilities using parallelism"""
        try:
            response = self.session.get(self.target_url, timeout=10)
            soup = BeautifulSoup(response.text, 'html.parser')
            forms = soup.find_all('form')
            
            print(f"[*] Found {len(forms)} forms to test for SQLi.")

            with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
                futures = [executor.submit(self._test_form_sqli, form, response.url) for form in forms]
                concurrent.futures.wait(futures)

            # Also test URL parameters
            self._test_url_sqli()

        except Exception as e:
            print(f"[-] Error testing SQLi: {e}")

    def _test_form_sqli(self, form, url):
        """Helper to test a single form for SQLi"""
        try:
            action = form.get('action')
            method = form.get('method', 'get').lower()
            inputs = form.find_all('input')
            
            target_url = urljoin(url, action)
            
            for payload in self.sqli_payloads:
                data = {}
                for input_tag in inputs:
                    if input_tag.get('type') in ['text', 'search', 'url', 'email', 'password']:
                        data[input_tag.get('name')] = payload
                    else:
                        data[input_tag.get('name')] = input_tag.get('value', '')
                
                try:
                    if method == 'post':
                        res = self.session.post(target_url, data=data, timeout=5)
                    else:
                        res = self.session.get(target_url, params=data, timeout=5)
                    
                    # Basic SQL error detection
                    errors = [
                        "SQL syntax", "mysql_fetch", "native client", "ORA-", 
                        "PostgreSQL query", "SQLite/JDBCDriver"
                    ]
                    
                    for error in errors:
                        if error.lower() in res.text.lower():
                            self.results.append({
                                'type': 'SQL Injection',
                                'severity': 'Critical',
                                'description': f"Possible SQL Injection found in form at {target_url} with payload {payload}",
                                'url': target_url,
                                'payload': payload
                            })
                            return
                except:
                    pass
        except:
             pass

    def _test_url_xss(self):
        """Test URL parameters for XSS"""
        parsed = urlparse(self.target_url)
        if not parsed.query:
            return
            
        params = parsed.query.split('&')
        base_url = self.target_url.split('?')[0]
        
        for payload in self.xss_payloads:
            for param in params:
                if '=' in param:
                    key = param.split('=')[0]
                    # Construct new query
                    new_query = f"{key}={payload}"
                    # Rebuild complete URL (simplified for single param injection)
                    target = f"{base_url}?{new_query}"
                    
                    try:
                        res = self.session.get(target, timeout=5)
                        if payload in res.text:
                            self.results.append({
                                'type': 'Reflected XSS (URL)',
                                'severity': 'High',
                                'description': f"XSS vulnerability found in URL parameter {key}",
                                'url': target,
                                'payload': payload
                            })
                    except:
                        pass

    def _test_url_sqli(self):
        """Test URL parameters for SQLi"""
        parsed = urlparse(self.target_url)
        if not parsed.query:
            return
            
        params = parsed.query.split('&')
        base_url = self.target_url.split('?')[0]
        
        for payload in self.sqli_payloads:
            for param in params:
                 if '=' in param:
                    key = param.split('=')[0]
                    new_query = f"{key}={payload}"
                    target = f"{base_url}?{new_query}"
                    
                    try:
                        res = self.session.get(target, timeout=5)
                        errors = ["SQL syntax", "mysql_fetch", "native client", "ORA-", "PostgreSQL query"]
                        for error in errors:
                            if error.lower() in res.text.lower():
                                 self.results.append({
                                    'type': 'SQL Injection (URL)',
                                    'severity': 'Critical',
                                    'description': f"SQL Injection found in URL parameter {key}",
                                    'url': target,
                                    'payload': payload
                                })
                    except:
                        pass
    
    def test_xss(self):
         # Legacy method wrapper
         self.test_xss_parallel()

    def test_sqli(self):
        # Legacy method wrapper
        self.test_sqli_parallel()

    def test_api_security(self):
        """Test API endpoints for security vulnerabilities"""
        # Common API paths
        api_paths = ['/api', '/rest', '/graphql',
                     '/v1', '/v2', '/swagger', '/openapi.json']

        for path in api_paths:
            url = urljoin(self.target_url, path)
            try:
                response = self.session.get(url, timeout=10)
                if response.status_code == 200:
                    # Test for API-specific vulnerabilities
                    if 'graphql' in path:
                        self.test_graphql_security(url)
                    elif 'swagger' in path or 'openapi' in path:
                        self.results.append({
                            'type': 'Exposed API Documentation',
                            'severity': 'Medium',
                            'description': f"API documentation is exposed at {url}",
                            'url': url
                        })
            except:
                pass

    def test_security_headers(self):
        """Check for missing security headers"""
        try:
            response = self.session.get(self.target_url, timeout=10)
            headers = response.headers

            security_headers = {
                'X-Frame-Options': 'Protects against Clickjacking attacks.',
                'Content-Security-Policy': 'Mitigates XSS and data injection attacks.',
                'Strict-Transport-Security': 'Enforces secure (HTTPS) connections.',
                'X-Content-Type-Options': 'Prevents MIME-sniffing.',
                'Referrer-Policy': 'Controls how much referrer information is included with requests.',
                'Permissions-Policy': 'Controls which browser features can be used.'
            }

            for header, description in security_headers.items():
                if header not in headers:
                    self.results.append({
                        'type': 'Missing Security Header',
                        'severity': 'Low', # Keep as Low/Medium to avoid alarm fatigue for common missing headers
                        'description': f"Missing {header} header. {description}",
                        'url': self.target_url
                    })
        except Exception as e:
            print(f"[-] Error checking security headers: {e}")

    def test_cookie_security(self):
        """Check for insecure cookie configurations"""
        try:
            response = self.session.get(self.target_url, timeout=10)
            cookies = response.cookies

            for cookie in cookies:
                issues = []
                if not cookie.secure:
                     issues.append("Missing 'Secure' flag")
                if not cookie.has_nonstandard_attr('HttpOnly') and not cookie.get_nonstandard_attr('HttpOnly'): # Requests cookie jar handling vary
                     # Check if it's actually HttpOnly (requests might not expose this easily depending on version, generic check)
                     # Inspecting raw headers is often more reliable for HttpOnly
                     pass 

                # Re-check via headers for Set-Cookie to be sure about flags
                pass
            
            # Robust check using Set-Cookie header parsing
            if 'Set-Cookie' in response.headers:
                # This catches multiple Set-Cookie headers
                # Note: requests merges them, but we can iterate if we access raw or splitting
                # For simplicity, we just look for flags in the string representation if possible or rely on the cookie jar
                pass

            # Improved Cookie Check iterating properly
            for cookie in self.session.cookies:
                 # Check Secure
                if not cookie.secure:
                    self.results.append({
                        'type': 'Insecure Cookie',
                        'severity': 'Medium',
                        'description': f"Cookie '{cookie.name}' is missing the 'Secure' flag.",
                        'url': self.target_url
                    })
                
                # Check HttpOnly (Requires some heuristic as requests.cookies doesn't always show it clearly on the object)
                # We can check the rest of the cookie attributes if available, but 'httponly' is often a specific attribute
                if not cookie.has_nonstandard_attr('HttpOnly'):
                     self.results.append({
                        'type': 'Insecure Cookie',
                        'severity': 'Medium',
                        'description': f"Cookie '{cookie.name}' may be missing the 'HttpOnly' flag.",
                        'url': self.target_url
                    })

        except Exception as e:
            print(f"[-] Error checking cookie security: {e}")



password = '@Vi3iq4y8'
email = 'bugtesting@gmail.com'
strings = ['red', 'bob', 'admin', 'alex', 'testing',
           'test', 'lol', 'yes', 'dragon', 'bad']  # list of generic, often-used words that may appear in usernames or weak passwords
# list of common variable or parameter names often found in web applications
commonNames = ['csrf', 'auth', 'token', 'verify', 'hash']
tokenPattern = r'^[\w\-_+=/]{14,256}$'  # regex token pattern

headers = {  # Custom Header
    'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
    'Accept-Language': 'en-US,en;q=0.5',
    'Accept-Encoding': 'gzip,deflate,br',  # 'br' is for Brotli compression
    'Connection': 'close',
    'DNT': '1',  # Do Not Track
    'Upgrade-Insecure-Requests': '1',
    # Custom User-Agent string
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:93.0) Gecko/20100101 Firefox/93.0',
    'Referer': 'https://www.example.com/',  # Referrer URL for context
    'Cache-Control': 'no-cache',  # Disable caching
    'Pragma': 'no-cache',  # Legacy header for cache control
    # MIME type for form submissions
    'Content-Type': 'application/x-www-form-urlencoded',
    # Used for API requests with token-based authentication
    'Authorization': 'Bearer <token>',
    'X-Requested-With': 'XMLHttpRequest',  # Indicates AJAX request
    'Origin': 'https://www.example.com',  # Originating site for CORS
    'Host': 'www.example.com',  # Host header to specify the domain being requested
    # Used for conditional GET requests
    'If-Modified-Since': 'Sat, 29 Oct 1994 19:43:31 GMT',
}
